function [H0_BirthTime, H0_DeathTime, H1_BirthTime, H1_DeathTime] = calculateBarcodes(gold, contourOrAll)

boundaryConnectivity = 'eight';
componentConnectivity = 4;
holeConnectivity = 8;

if strcmp(contourOrAll, 'contour')
    bnd = findContourMap(gold, boundaryConnectivity);
else
    bnd = gold > 0;
end
holes = imfill(bnd, holeConnectivity, 'holes');
holes = holes > 0 & bnd == 0;

[prevComp, prevCompNo] = bwlabeln(bnd, componentConnectivity);
H0_BirthTime = ones(prevCompNo, 1);
H0_DeathTime = zeros(prevCompNo, 1);

[prevHole, prevHoleNo] = bwlabeln(holes, holeConnectivity);
H1_BirthTime = ones(prevHoleNo, 1);
H1_DeathTime = zeros(prevHoleNo, 1);

currentTime = 2;
%while prevCompNo > 1 || prevHoleNo > 0
while sum(sum(bnd == 0)) > 0
    % Find connected components and holes in the current iteration
    bnd = increaseDistance(bnd);
    holes = imfill(bnd, holeConnectivity, 'holes');
    holes = holes > 0 & bnd == 0;

    % Update connected components (birth and death times)
    [currComp, currCompNo] = bwlabeln(bnd, componentConnectivity);
    overlaps = findOverlappingComponents(prevComp, currComp, prevCompNo, currCompNo);
    errorCheck4Components(overlaps, H0_DeathTime);
    [H0_BirthTime, H0_DeathTime, newIds] = updateBirthDeathTimes(overlaps, H0_BirthTime, H0_DeathTime, currentTime, 'component');
    currComp = relabelComponents(currComp, newIds);
    
    % Update holes (birth and death times)
    [currHole, currHoleNo] = bwlabeln(holes, holeConnectivity);
    overlaps = findOverlappingComponents(prevHole, currHole, prevHoleNo, currHoleNo);
    errorCheck4Holes(overlaps);
    [H1_BirthTime, H1_DeathTime, newIds] = updateBirthDeathTimes(overlaps, H1_BirthTime, H1_DeathTime, currentTime, 'hole');
    currHole = relabelComponents(currHole, newIds);
    
    prevHole = currHole;
    prevHoleNo = currHoleNo;
    
    prevComp = currComp;
    prevCompNo = currCompNo;
    currentTime = currentTime + 1;
    
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
% Thickens the boundary pixels by one towards their four neignbors     %
%----------------------------------------------------------------------%
function res = increaseDistance(bnd)

[dx, dy] = size(bnd);
res = bnd;
for i = 1 : dx
    for j = 1 : dy
        if bnd(i, j) == 0
            if i > 1 && bnd(i - 1, j) == 1
                res(i, j) = 1;
            elseif i < dx && bnd(i + 1, j) == 1
                res(i, j) = 1;
            elseif j > 1 && bnd(i, j - 1) == 1
                res(i, j) = 1;
            elseif j < dy && bnd(i, j + 1) == 1
                res(i, j) = 1;
            end
        end
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
% Finds overlapping components between previous and current iterations %
%----------------------------------------------------------------------%
function overlaps = findOverlappingComponents(prev, curr, prevNo, currNo)

overlaps = zeros(prevNo, currNo);
[dx, dy] = size(prev);
for i = 1 : dx
    for j = 1 : dy
        if prev(i, j) > 0 && curr(i, j) > 0
            overlaps(prev(i, j), curr(i, j)) = 1;
        end
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
function problemOccurs = errorCheck4Holes(overlaps)
    
problemOccurs = 0;
[prevNo, currNo] = size(overlaps);
for j = 1 : currNo
    cnt = 0;
    for i = 1 : prevNo
        if overlaps(i, j) == 1
            cnt = cnt + 1;
        end
    end
    if cnt > 1
        fprintf('Invalid death: more than one previous match for a new one\n');
        problemOccurs = 1;
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
function problemOccurs = errorCheck4Components(overlaps, deathTime)

problemOccurs = 0;
[prevNo, currNo] = size(overlaps);
% Check potential problems first
for i = 1 : prevNo
    cnt = 0;
    for j = 1 : currNo
        if overlaps(i, j) == 1
            cnt = cnt + 1;
        end
    end
    if cnt == 0 && deathTime(i) == 0
        fprintf('Invalid death: no current match for a previously existing one\n');
        problemOccurs = 1;
    end
    if cnt > 1
        fprintf('Invalid birth: more than one current match for a previously existing one\n');
        problemOccurs = 1;
    end
end
for j = 1 : currNo
    cnt = 0;
    for i = 1 : prevNo
        if overlaps(i, j) == 1
            cnt = cnt + 1;
        end
    end
    if cnt == 0
        fprintf('Invalid birth: no previous match for a new one\n');
        problemOccurs = 1;
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
function cmap = relabelComponents(cmap, newIds)

[dx, dy] = size(cmap);
for i = 1 : dx
    for j = 1 : dy
        if cmap(i, j)
            cmap(i, j) = newIds(cmap(i, j));
        end
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
function [birthTime, deathTime, newIds] = updateBirthDeathTimes(overlaps, birthTime, deathTime, currentTime, holeOrComponent)

[prevNo, currNo] = size(overlaps);
existingNo = size(birthTime, 1);
nextLabel = existingNo + 1;

newIds = zeros(currNo, 1);
for j = 1 : currNo
    cnt = 0;
    for i = 1 : prevNo
        if overlaps(i, j) == 1
            cnt = cnt + 1;
            if newIds(j, 1) == 0
                newIds(j, 1) = i;
            elseif strcmp(holeOrComponent, 'component')
                % previous components merge (many-to-one)
                % one component lives (if part), the others die
                deathTime(i, 1) = currentTime;
            end
        end
    end
    if cnt == 0 && strcmp(holeOrComponent, 'hole')
        % one hole newly borns (zero-to-one)
        newIds(j, 1) = nextLabel;
        birthTime(nextLabel, 1) = currentTime;
        deathTime(nextLabel, 1) = 0;
        nextLabel = nextLabel + 1;
    end
end

if strcmp(holeOrComponent, 'hole')
    for i = 1 : existingNo
        cnt = 0;
        if i <= prevNo
            for j = 1 : currNo
                if overlaps(i, j) == 1
                    cnt = cnt + 1;
                end
            end
        end
        % one existing hole dies (one-to-zero)
        if cnt == 0 && deathTime(i) == 0
            deathTime(i, 1) = currentTime;
        end
        if cnt > 0
            first = 1;
            for j = 1 : currNo
                if overlaps(i, j) == 1
                    if first == 1
                        newIds(j, 1) = i;
                        first = 0;
                    else
                        % first hole lives, the others newly born (one-to-many)
                        newIds(j, 1) = nextLabel;
                        birthTime(nextLabel, 1) = currentTime;
                        deathTime(nextLabel, 1) = 0;
                        nextLabel = nextLabel + 1;
                    end 
                end
            end
        end
    end
end

end
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
%----------------------------------------------------------------------%
